/*
 * Shutdown application
 *
 * (c) 2000 Draco/YC
 */

/*  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

# include <basepage.h>
# include <mintbind.h>
# include <string.h>
# include <fcntl.h>
# include <signal.h>
# include <gemma/gemma.h>

# include "xattr.h"
# include "shutdown.h"

# ifndef AP_AESTERM
#  define AP_AESTERM	52
# endif
# define STDOUT		1
# define STDERR		2
# define LSHIFT		0x0002

extern GEM_ARRAY *gem;		/* in libgemma.a, initialized at startup */
extern SLB gemma;		/* idem ibidem */

static unsigned long mode = 1;		/* warm boot is default */
static short nowait;
static short ok;
static const short buttons[] = { HALT, REBOOT, COLD, QUITGEM };

static inline void
aes_getinfo(short mode, short *ap)
{
	/* Notice that the real availability of this call is
	 * checked inside the library (no need to check the AES
	 * version number here)
	 */
	appl_getinfo(mode, &ap[0], &ap[1], &ap[2], &ap[3]);
}

static inline long
any_reply(unsigned long msec)
{
	short msg[8];
	long r;

	/* any reply or timeout is interesting */
	r = evnt_multi(MU_MESAG|MU_TIMER, 0, 0, 0, 0, 0, 0, 0, 0, \
			0, 0, 0, 0, 0, msg, msec, 0, 0, 0, 0, 0, 0);
	r = (r & MU_TIMER) ? 1 : 0;

	return r;
}

static void
initiate(WINDIAL *wd)
{
	short ref, pid, msg[8], ap[4];
	long r;

	aes_getinfo(AES_PROCESS, ap);

	bzero(msg, sizeof(msg));
	msg[1] = gem->global[2];

	for(;;)
	{
		ref = 0;
		if (shel_write(SWM_SHUTDOWN, SD_PARTIAL, 0, &ref, 0))
			break;
		if (!ref)
			return;
		msg[0] = AP_TERM;
		r = appl_write(ref, sizeof(msg), msg);
		if (r)
		{
			any_reply(3000UL);

			if (!ap[1])
			{
				pid = ref;
			}
			else
			{
				r = (long)ref;
				r &= 0x0000ffffL;
				r |= 0xfffe0000L;
				pid = appl_find((void *)r);
			}

			if (pid < 0)
				continue;

			(void)Pkill(pid, SIGINT);
			evnt_timer(1500L);
			r = Pkill(pid, SIGNULL);
			if (r < 0)
				continue;

			(void)Pkill(pid, SIGQUIT);
			evnt_timer(1500L);
			r = Pkill(pid, SIGNULL);
			if (r < 0)
				continue;

			(void)Pkill(pid, SIGTERM);
			evnt_timer(1500L);
			r = Pkill(pid, SIGNULL);
			if (r >= 0)
				(void)Pkill(pid, SIGKILL);
		}
	}

	do
	{
		evnt_mesag(msg);
	}
	while((msg[0] != SHUT_COMPLETED) && (msg[0] != AP_TERM));

	if (msg[0] == AP_TERM)
		return;

	/* Launch the GEM shutdown script
	 */
	r = proc_exec(100, 1, "/usr/local/etc/GEM/rc.shutdown", "", 0);
	if (r > 0)
		Pwaitpid(r, 0, 0);

	bzero(msg, sizeof(msg));

	/* Attempt to terminate the AES whenever possible */
	aes_getinfo(64, ap);
	if (ap[0])
	{
		Psigblock(1<<SIGTERM);
		msg[0] = AP_AESTERM;
		msg[1] = gem->global[2];
		shel_write(SWM_AESMSG, 0, 0, (void *)msg, 0);
	}
}

static inline void
load_inf(char *path)
{
	short file;
	unsigned long temp;
	long r;

	file = Fopen(path, O_RDONLY);
	if (file < 0)
		return;
	r = Fread(file, sizeof(temp), &temp);
	Fclose(file);
	if (r < sizeof(temp))		/* ... including < 0 */
		return;
	if (temp > 3)
		return;
	mode = temp;
}

static inline void
save_inf(char *path)
{
	short file;

	file = Fcreate(path, 0);
	if (file < 0)
		return;
	Fwrite(file, sizeof(mode), &mode);
	Fclose(file);
}

static inline void
deselect(WINDIAL *wd, short obj)
{
	OBJECT *ob = wd->wb_treeptr;

	if (!(ob[obj].ob_flags & OF_RBUTTON))
		objc_xchange(wd, obj, ob[obj].ob_state & ~OS_SELECTED, 1);
}

static long
newmsg(WINDIAL *wd, short vec, short *msg)
{
	if (msg[0] == AP_TERM)
		windial_longjmp(wd, vec);	/* never returns */
	return 0;
}

static void
do_window(WINDIAL *wd)
{
	OBJECT *obj = wd->wb_treeptr;
	short m = 0;

	windial_open(wd);

	if (windial_setjmp(wd, 0, newmsg) == 0)
	{
		do
		{
			m = windial_formdo(wd);
			if (m != -1)
			{
				m &= 0x7fff;
				deselect(wd, m);
			}
			switch(m)
			{
				case OK:
					objc_xchange(wd, OK, obj[OK].ob_state | OS_DISABLED, 1);
					objc_xchange(wd, CANCEL, obj[CANCEL].ob_state | OS_DISABLED, 1);
					windial_close(wd);
					initiate(wd);
					ok = 1;
					return;
				case REBOOT:
					mode = 1;
					break;
				case HALT:
					mode = 0;
					break;
				case COLD:
					mode = 2;
					break;
				case QUITGEM:
					mode = 3;
			}
		}
		while((m != CANCEL) && (m != -1));
	}
	windial_close(wd);
}

# if 0
static void
rts(void)
{}
# endif

int
main(void)
{
	WINDIAL *wd;
	OBJECT *obj;
	long r;
	short isroot, ap[4];
	char path[1024], *homepath;

	r = appl_open("shutdown.rsc", 0, PNAME);
	if (r < 0)
		return r;

	/* With latest lib this can't fail */
	wd = (WINDIAL *)windial_create(0, SHUTDOWN, ICON, 0, WINTITLE);

	homepath = env_get("HOME=");

	path[0] = 0;
	if (homepath)
	{
		strcpy(path, homepath);
		strcat(path, "\\");
	}
	strcat(path, "shutdown.inf");

	load_inf(path);

	isroot = Pgeteuid() ? 0 : 1;
	if (isroot)
	{
		char *sda;

		if ((Kbshift(-1) & LSHIFT) == 0)
		{
			sda = env_get("SDMASTER_ARGS=");
			if (sda)
			{
				r = strlen(sda);
				if (r >= sizeof("--nowait"))
				{
					while(*sda && (*sda == 0x20))
						sda++;
					if (*sda)
						if (strncmp(sda, "--nowait", sizeof("--nowait")) == 0)
							nowait = 1;
				}
			}
		}
	}

	obj = wd->wb_treeptr;
	if (!isroot)
	{
		obj[HALT].ob_state |= OS_DISABLED;
		obj[HALT].ob_flags &= ~OF_EXIT;
		obj[REBOOT].ob_state |= OS_DISABLED;
		obj[REBOOT].ob_flags &= ~OF_EXIT;
		obj[COLD].ob_state |= OS_DISABLED;
		obj[COLD].ob_flags &= ~OF_EXIT;
		mode = 3;
	}

	aes_getinfo(64, ap);
	if (!ap[0])
	{
		if (mode == 3)
			mode = 0;
		obj[QUITGEM].ob_state |= OS_DISABLED;
		obj[QUITGEM].ob_flags &= ~OF_EXIT;
		if (!isroot)
		{
			obj[OK].ob_state |= OS_DISABLED;
			obj[OK].ob_flags &= ~(OF_EXIT|OF_DEFAULT);
			obj[CANCEL].ob_flags |= OF_DEFAULT;
		}
	}

	if (!(obj[buttons[mode]].ob_state & OS_DISABLED))
		obj[buttons[mode]].ob_state |= OS_SELECTED;

	if (nowait)
		initiate(wd);
	else
		do_window(wd);

	windial_delete(wd);

	if (ok)
		save_inf(path);

	r = appl_close();

	if (ok && (mode < 3))
	{
		Sync();
//		if (!Pfork())	/* This causes a crash. Why? */
		{
# if 0
			Psignal(SIGALRM, rts);
			Tmalarm(3000L);
			Pause();
# endif
			Shutdown(mode);
		}
	}

	return r;
}

/* EOF */
